import { char } from './char.js'
import { fail, fork, mapStr } from './fn.js'
import { curry } from '../vendor/izuna/src/curry.js'

/** @import { ParserState } from './state.js' */

/**
 * @param {...any} parsers
 */
export const any = (...parsers) =>
  /**
   * @param {ParserState} state
   */
  state => {
    for (const parser of parsers) {
      const [original, clone] = fork(state)
      const result = parser(clone)
      if (result.isOk) {
        return result
      }
    }

    return fail('no matching parsers', state)
  }

export const anyOf = curry(
  /**
    * @param {string} str
    * @param {ParserState} state
    */
  (str, state) => (
    any(...mapStr(char, str))(state)
  ))

export const map = curry(
  /** 
   * @param {(...args: any[]) => any} fn
   * @param {Parser} parser
   * @param {ParserState} state
   */
  (fn, parser, state) => {
    return parser(state).chain(result => {
      try {
        /** @type {Result<ParserState, ParseError>} */
        const parsed = fn(diff(state[0], result[0]))
        const backtrack = Tuple(state[0], result[1])


        return succeed(parsed, backtrack)
      } catch (e) {
        return fail('failed to map', state, e)
      }
    })
  })

