import { ParseError } from './state.js'
import { fail, fork, succeed } from './fn.js'
import { anyChar } from './char.js'
import { ok } from '../vendor/kojima/src/index.js'
import { curry } from '../vendor/izuna/src/index.js'

/** @import { Result } from '../vendor/kojima/src/index.js' */
/** @import { ParserState } from './state.js' */

export const maybe = curry(
  /**
    * @param {(...args: any[]) => Result<ParserState, ParseError>} parser
    * @param {ParserState} state
    */
  (parser, state) => {
    const [original, clone] = fork(state)
    const result = parser(clone)
    return result.isOk() ? result : succeed([], original)
  })

export const not = curry((parser, state) => {
  const result = parser(state)

  if (result.isOk()) {
    return fail('"not" parser failed', state)
  } else {
    return succeed([], state)
  }
})

export const until = curry((parser, state) => {
  let result = ok(state)

  while (result.isOk()) {
    const [original, clone] = result.chain(fork)
    result = result.chain(x => parser(clone))
    if (result.isOk()) {
      break
    } else {
      result = anyChar(original)
    }
  }

  return result
})

export const skip = curry((parser, state) => {
})

